use std::sync::{Arc, Mutex};

use crate::metrics::Metrics;
use crate::{lox_context, DbConfig};
use chrono::{naive::Days, DateTime, Local, NaiveDateTime, Utc};
use lox_library::{BridgeAuth, BridgeDb};
use sled::IVec;

// Database of Lox Distributor State
pub struct DB {
    db: sled::Db,
}

impl DB {
    // Writes the Lox context to the lox database with "context_%Y-%m-%d_%H:%M:%S" as the
    // database key
    pub fn write_context(&mut self, context: lox_context::LoxServerContext) {
        let date = Local::now().format("context_%Y-%m-%d_%H:%M:%S").to_string();
        let json_result = serde_json::to_vec(&context).unwrap();
        println!("Writing context to the db with key: {:?}", date);
        let _new_ivec = self.db.insert(
            IVec::from(date.as_bytes().to_vec()),
            IVec::from(json_result.clone()),
        );
        assert_eq!(
            self.db
                .get(IVec::from(date.as_bytes().to_vec()))
                .unwrap()
                .unwrap(),
            IVec::from(json_result)
        );
    }

    // If roll_back_date is empty, opens the most recent entry in the lox database or if none exists, creates a
    // new database. If roll_back_date is not empty, use the specified date to roll back to a previous lox-context
    // either exactly the entry at the roll_back_date or within 24 hours from thhe roll_back_date.
    pub fn open_new_or_existing_db(
        db_config: DbConfig,
        roll_back_date: Option<String>,
        metrics: Metrics,
    ) -> Result<(DB, lox_context::LoxServerContext), sled::Error> {
        let mut context: lox_context::LoxServerContext;
        let (lox_db, context) = match sled::open(db_config.db_path) {
            Ok(lox_db) => {
                // Check if the lox_db already exists
                if lox_db.was_recovered() {
                    context = read_lox_context_from_db(lox_db.clone(), roll_back_date);
                    context.metrics = metrics;
                //Otherwise, create a new Lox context
                } else {
                    let new_db = BridgeDb::new();
                    let new_ba = BridgeAuth::new(new_db.pubkey);
                    context = lox_context::LoxServerContext {
                        db: Arc::new(Mutex::new(new_db)),
                        ba: Arc::new(Mutex::new(new_ba)),
                        extra_bridges: Arc::new(Mutex::new(Vec::new())),
                        to_be_replaced_bridges: Arc::new(Mutex::new(Vec::new())),
                        metrics,
                    };
                }
                (DB { db: lox_db }, context)
            }
            Err(e) => {
                panic!("Unable to read or create lox database! {:?}", e);
            }
        };
        Ok((lox_db, context))
    }
}

// Logic for finding the correct context to open from the database
fn read_lox_context_from_db(
    lox_db: sled::Db,
    roll_back_date: Option<String>,
) -> lox_context::LoxServerContext {
    let context: lox_context::LoxServerContext;
    // Check if there is a roll back date and try to choose the appropriate context
    // to rollback to, otherwise, take the last saved context
    match roll_back_date {
        // If roll back date has been specified, either the exact date or range should be set
        Some(roll_back_date) => {
            // If the date is specified and it's in the database, use that to populate the context
            if lox_db.contains_key(roll_back_date.clone()).unwrap() {
                // Find date/time in db and use the context from that date.
                let ivec_context = lox_db
                    .get(IVec::from(roll_back_date.as_bytes().to_vec()))
                    .unwrap()
                    .unwrap();
                context = serde_json::from_slice(&ivec_context).unwrap();
                println!("Successfully used exact key {:?}", roll_back_date);
            } else {
                // If the exact date is not found, compute the start of the range as 24 hours prior to the specified date
                match compute_startdate_string(roll_back_date.clone()){
                    Some(start_date) => {
                        let start_date = start_date.format("context_%Y-%m-%d_%H:%M:%S").to_string();
                        let r: sled::Iter = lox_db.range(IVec::from(start_date.as_bytes().to_vec())..IVec::from(roll_back_date.as_bytes().to_vec()), );
                        match r.last(){
                            Some(entry) => {
                                let ivec_context = entry.unwrap();
                                let key: String = String::from_utf8(ivec_context.0.to_vec()).unwrap();
                                println!("Successfully used range and key {:?}", key);
                                context = serde_json::from_slice(&ivec_context.1).unwrap();
                            },
                            //If no entries are found within the 24 hours prior to the specified date, Exit
                            None => panic!("UNEXPECTED DATE: No entries found in the 24 hours prior to the input roll_back_date"),
                        }
                    }
                    None => panic!("UNEXPECTED DATE: Tried to use input date to calculate 24 hour range but failed. Exiting."),
                }
            }
        }
        // Use the last entry to populate the Lox context if no rollback date is set (which should be most common)
        None => {
            context = use_last_context(lox_db);
        }
    }
    context
}

fn compute_startdate_string(date_range_end: String) -> Option<DateTime<Utc>> {
    let parsed_end =
        NaiveDateTime::parse_from_str(&date_range_end, "context_%Y-%m-%d_%H:%M:%S").unwrap();
    let dt = DateTime::<Utc>::from_naive_utc_and_offset(parsed_end, Utc);
    dt.with_timezone(&Utc).checked_sub_days(Days::new(1))
}

// Use the last context that was entered into the database
fn use_last_context(lox_db: sled::Db) -> lox_context::LoxServerContext {
    let ivec_context = lox_db.last().unwrap().unwrap();
    let ivec_date: String = String::from_utf8(ivec_context.0.to_vec()).unwrap();
    println!("Using last context with date: {:?}", ivec_date);
    serde_json::from_slice(&ivec_context.1).unwrap()
}

#[cfg(test)]
mod tests {
    use super::lox_context::LoxServerContext;
    use super::DbConfig;
    use super::Metrics;
    use super::DB;

    #[test]
    fn test_write_context() {
        let (mut lox_db, _context) =
            DB::open_new_or_existing_db(DbConfig::default(), None, Metrics::default()).unwrap();
        assert!(
            lox_db.db.is_empty(),
            "db read from context that shouldn't exist"
        );
        let test_string = "{\"db\":{\"keypair\":[121,4,160,52,248,243,39,240,71,101,82,207,64,22,142,74,246,212,153,208,24,125,84,1,155,176,83,205,46,185,185,44],\"pubkey\":[160,195,123,233,25,157,126,89,37,85,193,202,132,49,53,205,169,199,176,14,251,234,40,100,156,84,155,125,186,244,134,220],\"openinv_buckets\":[5,4,7,3,8],\"distributed_buckets\":[9],\"current_k\":2,\"daily_bridges_distributed\":1},\"ba\":{\"lox_priv\":{\"x0tilde\":[143,9,251,99,139,233,249,99,12,151,209,199,61,9,226,112,162,129,137,136,195,3,66,179,223,161,50,85,30,157,37,0],\"x\":[[243,220,21,212,62,251,120,189,78,69,185,186,29,7,23,200,129,129,142,47,138,112,151,4,126,40,138,144,220,88,86,8],[79,133,82,56,114,75,230,209,153,126,84,177,85,238,191,235,7,187,196,196,228,60,103,178,44,88,5,143,238,120,72,12],[114,190,197,197,92,27,104,48,174,133,173,50,91,173,81,40,202,254,102,100,238,9,180,244,14,47,71,72,202,169,223,0],[109,226,189,23,245,119,6,254,111,241,15,1,66,129,65,190,153,106,160,142,102,92,76,111,147,235,77,118,198,131,78,8],[219,73,214,47,104,208,140,251,118,150,220,157,85,115,83,86,89,232,171,242,152,64,20,61,209,119,78,245,36,204,85,11],[126,164,101,171,7,207,113,121,133,92,210,120,97,92,70,89,142,29,89,58,9,83,90,126,153,202,216,250,123,157,30,5],[25,203,47,124,192,229,50,210,39,186,172,157,71,221,90,252,83,0,234,180,15,228,133,110,58,222,51,147,110,190,237,5]]},\"lox_pub\":{\"X\":[[52,135,83,91,17,4,67,24,173,18,214,24,58,154,63,119,67,136,188,179,157,120,218,195,200,188,209,141,93,24,124,94],[238,46,17,216,198,220,35,70,61,27,186,253,177,7,221,231,169,81,99,102,18,128,125,114,184,15,246,233,1,244,118,61],[202,55,212,100,56,227,184,106,93,175,150,254,8,19,166,25,194,60,224,172,26,76,98,122,17,253,72,94,54,207,109,21],[216,173,4,2,65,255,161,45,106,111,132,219,209,220,33,26,43,173,120,117,116,30,45,151,244,206,240,255,231,179,175,103],[150,200,66,157,230,92,163,117,225,136,193,109,1,35,120,151,48,182,82,35,7,93,239,212,39,67,11,15,239,235,234,13],[172,116,53,192,167,34,113,160,85,68,184,65,117,63,29,26,138,132,224,234,105,96,194,93,220,240,88,237,110,113,86,82],[18,225,205,220,116,77,165,132,231,244,189,5,159,221,110,177,187,104,168,78,255,134,92,17,109,161,77,58,72,192,205,114]]},\"migration_priv\":{\"x0tilde\":[211,143,194,8,37,8,12,55,39,248,78,75,118,168,145,205,240,248,252,53,185,227,9,219,69,45,75,103,105,146,187,7],\"x\":[[104,107,78,46,137,189,128,168,148,77,184,118,199,234,137,81,190,103,120,180,134,231,215,197,20,176,51,138,238,54,110,12],[31,250,96,78,75,51,84,243,27,66,21,169,30,15,13,96,37,74,133,201,120,159,98,129,42,239,221,169,67,134,143,8],[74,167,255,153,36,234,214,227,250,25,106,172,36,25,72,176,81,222,197,223,237,86,17,230,97,249,107,70,91,21,190,15],[23,236,100,248,42,74,41,2,155,96,226,199,117,227,182,36,96,99,177,192,222,251,105,188,196,163,43,80,49,30,22,10],[242,31,173,104,138,50,162,156,255,85,58,180,93,228,130,239,115,14,89,179,114,236,140,190,116,16,6,147,121,27,92,11]]},\"migration_pub\":{\"X\":[[66,150,237,69,9,213,22,190,254,2,95,43,154,115,131,141,71,35,197,202,154,163,159,164,64,16,248,83,214,60,227,23],[26,26,30,134,61,140,209,165,221,146,165,203,213,57,60,16,82,185,177,218,28,59,86,37,139,88,118,69,43,120,49,97],[142,238,183,118,107,13,48,154,94,204,119,104,218,179,149,208,242,151,228,117,6,125,138,65,92,160,91,166,62,75,224,127],[6,134,209,57,107,68,65,24,78,50,197,79,39,74,239,181,77,57,17,173,30,21,231,241,211,185,78,235,219,138,235,49],[88,146,145,147,186,255,230,241,145,139,149,123,247,101,222,173,86,23,22,48,230,0,148,222,205,143,17,5,5,63,199,55]]},\"migrationkey_priv\":{\"x0tilde\":[108,193,14,7,211,244,9,190,251,186,173,191,159,67,226,168,60,109,225,13,157,225,5,88,42,21,40,103,143,79,232,0],\"x\":[[208,84,202,133,58,81,191,86,34,124,80,151,105,145,135,13,250,255,151,217,162,212,204,212,244,215,202,138,14,93,6,11],[253,152,102,66,182,93,38,129,126,173,154,32,46,2,96,174,197,211,100,130,19,206,64,153,40,104,51,36,99,163,154,10],[32,158,137,175,136,206,72,252,69,225,103,80,1,175,238,240,204,210,185,1,85,206,146,234,252,201,193,19,150,76,254,9]]},\"migrationkey_pub\":{\"X\":[[158,136,224,166,235,252,254,136,145,186,93,50,217,48,51,242,90,47,99,179,211,206,102,222,15,205,61,20,178,224,11,85],[238,183,52,40,190,24,231,89,149,158,183,90,255,178,156,37,47,223,162,251,128,222,16,185,125,91,29,185,76,121,209,15],[62,76,165,228,36,162,30,208,29,173,174,40,129,38,182,84,196,210,67,125,115,64,85,197,118,114,39,52,111,82,125,13]]},\"reachability_priv\":{\"x0tilde\":[36,187,28,100,96,130,72,135,158,51,170,109,52,72,103,90,220,74,212,104,103,92,247,69,35,235,11,53,107,234,77,0],\"x\":[[199,24,219,64,123,139,129,139,190,139,114,172,254,52,193,46,65,66,87,240,7,10,139,158,232,40,99,230,210,179,71,5],[186,173,85,159,209,12,113,8,136,170,194,128,243,174,20,78,188,57,63,144,163,218,59,102,107,184,152,42,149,22,66,6],[110,162,248,148,212,200,180,133,86,76,209,152,185,64,46,128,131,95,13,130,190,151,83,24,102,96,201,96,179,133,48,3]]},\"reachability_pub\":{\"X\":[[70,56,136,60,123,40,11,124,4,170,204,160,93,221,48,117,145,66,146,206,162,54,24,98,189,115,191,102,252,51,83,110],[60,52,219,76,2,245,192,204,150,195,154,211,150,104,201,53,43,137,32,135,133,205,82,109,50,186,87,52,27,254,160,46],[168,172,24,38,194,113,8,1,124,179,17,190,244,114,228,122,53,24,62,138,27,177,89,238,167,79,64,104,39,10,216,94]]},\"invitation_priv\":{\"x0tilde\":[98,218,42,226,106,65,201,104,75,104,13,80,9,213,177,167,186,16,188,139,208,56,23,22,16,17,206,187,170,253,122,15],\"x\":[[112,9,221,148,24,75,246,241,128,87,120,9,146,203,160,238,250,10,215,116,25,138,172,150,74,244,19,67,248,116,234,5],[246,214,228,161,154,83,161,139,113,170,189,7,6,84,104,101,167,201,174,249,29,219,155,108,165,19,88,252,32,169,121,14],[115,155,88,53,123,252,230,51,173,10,205,187,83,108,129,241,144,178,210,189,22,197,204,174,164,143,125,101,82,139,18,1],[235,162,227,81,108,197,23,213,19,74,249,242,84,12,11,134,244,102,171,74,125,234,94,134,17,200,235,172,3,236,212,9],[178,57,128,236,178,248,50,5,214,239,97,153,20,55,197,220,217,98,4,118,246,131,124,90,113,121,169,129,211,175,216,1]]},\"invitation_pub\":{\"X\":[[18,91,111,4,165,234,85,12,151,195,5,213,91,191,167,191,89,60,218,90,204,99,35,166,37,73,105,134,227,169,158,27],[52,2,100,3,112,118,66,31,105,121,74,208,215,223,40,210,103,149,254,18,232,150,136,241,112,136,4,98,187,175,214,25],[156,112,201,126,204,98,91,120,114,189,170,32,238,249,248,163,196,109,205,150,163,209,49,95,193,113,191,226,255,83,0,86],[120,15,100,23,59,173,212,124,73,190,37,77,132,190,169,48,183,222,194,183,119,194,39,184,73,206,193,154,200,129,252,117],[78,246,199,66,152,244,85,179,40,77,215,77,81,206,248,236,102,112,244,74,247,180,207,195,109,70,95,242,207,58,179,109]]},\"bridgedb_pub\":[160,195,123,233,25,157,126,89,37,85,193,202,132,49,53,205,169,199,176,14,251,234,40,100,156,84,155,125,186,244,134,220],\"bridge_table\":{\"counter\":10,\"keys\":{\"4\":[98,77,88,28,245,60,12,162,231,142,175,244,134,36,108,189],\"5\":[231,37,133,52,229,253,115,13,11,167,181,65,108,249,202,39],\"8\":[104,79,30,48,46,140,65,219,127,227,243,161,132,163,136,8],\"10\":[156,165,205,136,12,139,204,191,14,8,129,174,126,95,179,80],\"3\":[151,150,241,156,53,136,84,201,6,109,247,5,174,211,73,255],\"1\":[111,79,32,173,35,142,62,131,8,17,215,22,0,36,146,65],\"6\":[40,76,205,122,89,219,86,59,189,63,21,108,107,149,34,120],\"9\":[138,27,80,18,27,19,245,215,20,19,26,158,17,7,77,194],\"7\":[158,57,44,39,78,80,130,81,160,117,234,188,89,51,62,134],\"2\":[108,105,146,211,217,191,102,209,15,44,87,22,53,35,34,243]},\"buckets\":{\"2\":[{\"addr\":[49,57,53,46,55,50,46,55,52,46,49,49,55,0,0,0],\"port\":28746,\"uid_fingerprint\":8013544061081454931,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,50,57,49,49,48,66,55,68,57,69,57,54,48,65,53,67,52,50,56,52,56,55,55,50,65,70,54,50,52,65,51,66,50,55,65,50,68,69,52,69,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[49,50,56,46,49,51,57,46,52,52,46,49,53,55,0,0],\"port\":10839,\"uid_fingerprint\":16072135124933872676,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,55,67,53,65,51,68,67,53,51,49,48,55,53,53,48,54,65,50,52,52,65,51,67,70,49,57,66,53,55,70,52,68,69,52,55,55,67,65,54,51,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[49,51,52,46,50,48,55,46,50,51,53,46,49,48,53,0],\"port\":9042,\"uid_fingerprint\":5838789323103363859,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,56,48,55,56,70,57,55,66,50,50,49,53,55,69,66,55,49,54,65,69,66,68,53,54,65,57,52,55,54,67,51,55,56,57,57,65,68,70,54,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"9\":[{\"addr\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\"port\":63174,\"uid_fingerprint\":9189545078772697644,\"info\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,69,68,67,51,65,65,48,51,70,69,54,69,53,50,66,69,55,55,56,68,68,52,54,70,50,55,68,51,53,56,57,57,51,66,55,65,49,48,65,52,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"7\":[{\"addr\":[49,57,49,46,49,56,48,46,49,48,51,46,50,51,54,0],\"port\":50309,\"uid_fingerprint\":17576062521251310994,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"6\":[{\"addr\":[49,57,49,46,49,56,48,46,49,48,51,46,50,51,54,0],\"port\":50309,\"uid_fingerprint\":17576062521251310994,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\"port\":63174,\"uid_fingerprint\":922068108643293997,\"info\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]},{\"addr\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\"port\":63174,\"uid_fingerprint\":9189545078772697644,\"info\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,69,68,67,51,65,65,48,51,70,69,54,69,53,50,66,69,55,55,56,68,68,52,54,70,50,55,68,51,53,56,57,57,51,66,55,65,49,48,65,52,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]}],\"4\":[{\"addr\":[49,50,56,46,49,51,57,46,52,52,46,49,53,55,0,0],\"port\":10839,\"uid_fingerprint\":16072135124933872676,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,55,67,53,65,51,68,67,53,51,49,48,55,53,53,48,54,65,50,52,52,65,51,67,70,49,57,66,53,55,70,52,68,69,52,55,55,67,65,54,51,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"5\":[{\"addr\":[49,51,52,46,50,48,55,46,50,51,53,46,49,48,53,0],\"port\":9042,\"uid_fingerprint\":5838789323103363859,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,56,48,55,56,70,57,55,66,50,50,49,53,55,69,66,55,49,54,65,69,66,68,53,54,65,57,52,55,54,67,51,55,56,57,57,65,68,70,54,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"10\":[{\"addr\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\"port\":63174,\"uid_fingerprint\":14821306939685929169,\"info\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,66,48,56,49,49,56,65,56,48,68,55,54,54,66,65,57,65,56,70,57,65,55,50,50,67,50,54,54,49,52,69,48,54,66,56,54,69,52,67,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]},{\"addr\":[57,55,46,49,55,53,46,57,49,46,55,0,0,0,0,0],\"port\":27719,\"uid_fingerprint\":9270304952963857555,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,69,68,67,51,65,65,48,51,70,69,54,69,53,50,66,69,55,55,56,68,68,52,54,70,50,55,68,51,53,56,57,57,51,66,55,65,49,48,65,52,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[50,50,46,50,52,49,46,56,49,46,54,50,0,0,0,0],\"port\":36457,\"uid_fingerprint\":14409951350199027285,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,68,68,56,56,57,53,48,56,48,48,50,70,57,50,53,67,70,55,52,69,49,51,48,68,49,48,49,67,68,70,49,67,49,51,51,65,65,65,68,49,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"3\":[{\"addr\":[49,57,53,46,55,50,46,55,52,46,49,49,55,0,0,0],\"port\":28746,\"uid_fingerprint\":8013544061081454931,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,50,57,49,49,48,66,55,68,57,69,57,54,48,65,53,67,52,50,56,52,56,55,55,50,65,70,54,50,52,65,51,66,50,55,65,50,68,69,52,69,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"8\":[{\"addr\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\"port\":63174,\"uid_fingerprint\":922068108643293997,\"info\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],\"port\":0,\"uid_fingerprint\":0,\"info\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}],\"1\":[{\"addr\":[50,49,50,46,49,54,54,46,49,54,54,46,54,55,0,0],\"port\":32414,\"uid_fingerprint\":3643046579853129326,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,66,48,56,49,49,56,65,56,48,68,55,54,54,66,65,57,65,56,70,57,65,55,50,50,67,50,54,54,49,52,69,48,54,66,56,54,69,52,67,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[49,51,50,46,49,56,49,46,49,53,53,46,55,54,0,0],\"port\":55241,\"uid_fingerprint\":4215620865531854407,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,67,66,70,56,51,51,51,49,54,53,70,49,67,51,53,65,68,70,48,54,66,70,52,50,53,49,48,52,69,56,51,51,65,68,68,65,49,54,53,66,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},{\"addr\":[49,53,54,46,55,50,46,50,49,57,46,52,57,0,0,0],\"port\":56030,\"uid_fingerprint\":8564141605929456761,\"info\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,68,56,66,67,67,48,69,67,70,69,50,67,57,54,53,65,66,67,50,57,67,54,50,55,70,68,67,69,67,48,66,48,54,67,53,65,68,66,49,54,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}]},\"encbuckets\":{\"5\":[212,104,142,74,232,164,167,215,241,173,77,228,245,132,164,225,11,40,120,21,101,201,141,172,165,1,169,6,103,29,201,195,239,64,154,123,117,201,31,58,194,198,58,27,147,125,3,99,188,202,0,92,144,102,139,27,250,91,206,233,232,25,130,91,199,92,186,150,33,0,168,1,133,229,208,96,104,225,211,42,136,126,37,209,110,149,128,66,145,196,77,148,55,38,221,115,205,184,149,69,34,137,39,109,135,58,85,10,134,181,86,223,200,32,90,195,75,203,25,123,211,183,22,232,122,50,105,151,39,29,185,135,37,164,228,102,249,177,158,27,186,128,59,3,28,120,156,8,222,247,100,213,144,61,31,57,253,72,62,202,216,137,226,217,46,136,135,169,151,216,101,95,210,43,158,172,11,38,45,166,91,240,116,102,84,198,129,255,84,114,82,235,162,163,25,206,227,66,88,161,170,255,46,43,243,216,126,253,116,246,136,201,13,127,108,121,249,188,119,223,9,62,216,38,152,45,119,159,174,229,255,52,36,97,11,112,137,43,147,85,104,160,128,205,8,135,236,211,78,137,237,9,130,196,17,5,235,7,116,184,46,161,0,138,234,15,120,169,128,30,143,37,161,161,7,8,245,130,192,173,239,38,171,77,88,139,227,41,162,87,111,165,218,3,102,194,231,208,167,181,252,199,82,79,76,215,82,49,114,93,80,144,111,96,153,145,128,161,173,148,223,151,43,198,82,255,170,10,89,43,203,137,200,70,113,150,115,209,185,128,76,134,81,163,104,231,134,113,9,89,206,155,162,216,44,145,238,160,112,216,162,143,189,38,56,218,188,224,17,249,174,169,174,62,35,56,110,10,189,85,43,232,96,177,88,23,151,73,31,15,107,158,186,21,112,249,236,58,111,126,11,159,138,120,22,213,83,9,129,181,148,5,180,210,82,203,248,182,254,98,212,175,154,155,20,232,34,158,127,223,76,253,105,145,132,104,180,205,59,188,223,213,86,25,151,241,176,167,116,164,14,196,79,17,192,162,76,189,57,48,103,196,114,248,32,61,77,144,14,200,134,213,13,190,103,179,90,96,174,43,217,229,187,64,193,128,134,41,207,112,41,235,144,5,219,43,147,41,191,138,89,208,30,81,184,2,31,166,121,159,211,242,80,164,3,227,146,11,248,172,198,32,177,89,23,115,55,167,172,127,75,93,185,213,115,70,134,226,26,117,53,238,187,3,147,43,20,142,56,148,19,170,219,107,128,0,117,149,172,40,236,186,250,102,242,208,154,101,159,110,162,64,113,203,113,5,209,79,91,180,75,99,37,13,10,207,41,199,239,51,132,237,216,35,30,143,232,229,252,152,92,149,76,36,12,241,113,110,177,98,226,147,173,90,21,68,106,218,85,242,27,22,213,79,115,24,191,247,80,94,223,193,77,70,25,204,57,204,81,25,39,116,32,162,60,192,239,134,162,233,186,13,84,200,111,142,150,67,229,246,48,119,129,113,39,200,254,104,97,194,62,108,103,220,120,221,138,5,4,191,89,164,144,243,175,75,119,127,100,90,121,140,33,223,21,204,141,79,168,218,102,58,169,107,108,23,215,228,237,77,162,142,232,29,224,192,74,227,1,208,26,170,156,183,134,69,132,159,128,249,105,215,72,138,87,181,164,96,243,3,130,94,90,89,122,196,27,237,132,203,91,29,64,67,120,98,139,181,62,252,39,67,120,34,84,55,38,74,218,170,113,115,255,140,30,207,8,244,250,111,82,243,19,107,80,40,120,209,103,239,36,88,72,151,89,172,12,194,184,42,126,110,49,90,66,251,222,70,224,216,9,211,104,254,193,212,169,241,27,64,198,147,80,100,106,46,143,171,131,34,91,116,231,10,14,73,113,12,255,71,243,232,123,30,226,203,172,3,31,148,171,37,210,251,17,0,87,244,206,176,64,39,80,125,67,33,24,253,38,250,4,156,147,175,132,4,7,208,169,251,137,214,139,212,150,43,75,39,98,180,173,96,150,198,54,39,217,229,171,206,139,57,239,1,210,155,164,18,114,207,57,103,63,95,85,77,201,1,155,190,3,51,177,75,145,250,110,106,87,65,28,210,40,73,155,176,82,252,230,239,51,87,195,5,173,37,54,234,106,201,212,43,202,177,19,217,109,179,210,12,1,222,251,167,213,6,244,80,224,239,67,124,123,253,220,73,194,165,191,61,225,86,142,123,103,231,248,181,117,137,92,221,50,214,78,112],\"7\":[133,67,45,40,43,226,44,44,55,133,123,44,38,130,99,59,186,45,165,90,12,117,245,92,142,58,66,87,125,251,210,214,132,228,64,30,149,114,212,247,217,223,241,174,38,63,95,97,164,250,225,169,70,152,184,254,247,24,51,143,240,246,213,237,161,101,70,47,173,234,57,79,56,160,210,81,230,73,42,198,255,109,176,54,211,15,124,40,125,88,249,72,193,51,98,248,246,153,137,171,157,197,57,252,163,83,132,122,127,52,210,116,83,240,126,17,71,77,244,249,25,213,183,86,245,158,214,98,101,230,95,96,74,4,166,90,141,77,200,78,108,132,153,162,133,59,177,106,38,158,223,133,207,51,219,198,241,192,244,167,162,176,1,13,191,181,235,103,40,227,183,60,15,77,122,154,120,94,249,23,53,128,232,62,31,84,106,178,105,36,143,66,254,120,219,128,206,58,207,20,95,134,80,3,176,130,67,215,42,2,170,43,141,151,133,83,5,254,187,96,102,111,181,195,71,25,163,203,227,33,44,197,5,205,224,125,205,240,192,93,178,115,16,117,249,186,229,144,174,60,60,197,137,144,77,154,63,169,198,144,71,251,238,9,62,179,140,187,203,58,112,214,26,216,235,41,112,173,131,108,12,185,168,93,35,96,110,127,108,205,190,237,61,173,93,187,165,214,166,157,197,20,191,156,5,217,246,115,216,76,112,85,107,243,111,153,255,148,229,16,162,82,60,86,2,199,64,149,23,23,124,196,193,186,75,114,204,30,197,82,253,237,141,204,246,136,235,255,29,250,203,194,32,207,202,14,26,149,193,230,35,89,8,170,107,45,77,223,66,76,248,76,156,127,149,126,58,224,209,21,43,133,135,140,67,231,101,86,177,78,252,184,153,36,107,239,171,38,236,101,172,82,144,155,81,248,18,60,190,94,101,171,39,11,191,246,152,149,164,26,80,165,58,254,136,11,221,226,159,148,15,217,173,176,82,221,182,99,187,169,94,80,177,66,63,220,234,25,138,27,208,193,3,165,219,14,42,12,249,86,72,132,201,18,139,226,93,248,209,232,106,15,178,164,3,50,226,140,171,86,101,206,220,55,239,255,211,43,41,82,166,168,43,101,170,228,87,55,156,164,28,131,107,33,23,174,156,132,219,129,128,137,33,21,193,205,64,28,195,242,173,38,191,193,181,64,13,163,112,104,28,50,142,153,54,70,170,8,114,69,144,2,208,152,30,158,167,15,161,94,87,148,178,186,205,16,139,158,174,111,148,51,64,210,190,88,226,206,99,26,226,247,147,84,181,29,75,125,217,238,55,191,39,148,43,41,47,143,198,38,128,173,59,102,13,15,89,59,43,158,201,200,15,56,51,38,109,165,77,40,167,119,122,87,96,191,40,130,186,79,251,161,10,227,69,19,46,235,51,104,7,253,217,1,32,59,187,210,147,92,204,109,82,226,202,71,120,196,146,31,85,230,55,99,157,42,68,217,209,112,149,19,225,34,72,197,187,147,195,122,187,43,105,60,36,134,112,59,127,171,55,189,27,87,240,32,92,134,138,95,162,47,214,197,157,103,49,249,173,13,25,233,75,200,162,4,180,15,44,15,210,101,199,47,197,40,67,22,78,232,230,92,2,113,209,177,199,105,217,5,133,57,164,182,8,43,113,151,253,159,92,96,212,208,201,69,139,224,97,38,28,204,46,249,111,157,74,127,42,159,159,194,175,246,142,70,252,226,236,99,98,110,131,25,254,60,207,31,204,51,24,28,164,157,3,202,100,215,12,49,33,121,202,160,185,71,204,188,187,144,93,10,212,104,80,93,6,11,217,164,157,86,36,250,111,226,207,234,43,58,1,229,42,178,37,155,144,1,179,188,38,50,121,239,241,124,80,117,191,115,100,30,196,58,66,39,2,138,220,251,12,83,37,3,108,22,115,232,75,25,99,168,19,190,27,22,79,18,187,190,138,93,237,41,28,199,99,30,125,169,230,49,2,169,9,245,141,232,152,171,26,189,152,146,165,230,134,238,21,195,92,60,115,139,148,71,32,48,186,236,155,164,234,107,211,204,77,76,59,231,151,75,230,156,84,171,151,146,126,181,9,212,152,123,25,26,87,29,250,55,204,121,221,237,16,58,160,219,54,12,223,36,147,157,94,192,45,185,13,32,214,112,155,242,245,62,46,96,29,236,16,2,54,28,229,9,35,7,11,110,251,181,81,44,136,35,75,93,50,168,15,64,217,236,246,156],\"9\":[32,9,165,173,194,126,208,224,139,145,74,38,105,240,179,139,109,250,199,118,91,133,178,129,211,139,235,127,167,187,206,12,57,15,143,58,198,104,64,245,15,93,27,93,17,60,3,188,121,220,93,161,67,112,235,60,25,72,86,253,141,79,214,233,96,26,98,108,108,208,205,34,16,114,169,239,4,19,1,151,57,204,20,182,113,161,187,22,65,224,94,179,23,122,218,224,162,17,111,73,122,64,111,9,124,238,189,223,64,109,205,150,105,193,213,190,173,22,27,139,191,31,46,190,212,86,101,104,13,13,94,21,153,35,232,171,223,141,119,85,185,142,118,137,166,143,129,249,54,48,66,111,90,152,147,54,120,196,58,22,74,234,81,36,44,95,163,67,76,124,44,149,141,19,159,77,186,53,25,161,15,117,10,137,64,182,127,73,20,175,197,77,116,113,148,135,73,204,124,121,141,38,46,228,20,188,253,190,165,40,183,126,210,82,198,55,50,225,84,168,187,97,6,56,77,246,195,211,84,38,86,241,165,244,11,244,252,152,238,154,184,165,77,185,32,122,68,70,167,245,211,247,189,75,220,93,240,199,245,139,16,218,5,157,87,180,67,70,223,201,5,91,102,161,111,197,22,201,237,121,111,72,138,237,80,249,205,149,249,202,183,42,26,33,7,115,166,103,152,97,182,68,200,203,95,238,20,113,206,96,203,23,78,58,68,136,84,57,118,155,82,101,161,146,8,81,126,77,85,187,82,20,181,183,91,103,19,249,77,7,150,4,242,6,45,46,23,210,34,32,228,215,68,78,191,174,37,129,36,102,219,187,154,90,124,147,186,191,241,116,76,137,96,151,243,192,90,97,13,88,170,114,252,80,41,238,145,113,205,167,131,161,251,16,111,132,215,215,31,197,130,88,3,152,75,50,111,148,104,45,211,69,206,216,80,167,69,109,30,37,94,33,50,233,51,191,65,48,169,89,127,142,123,79,84,176,17,150,187,30,114,214,204,220,180,180,61,33,156,244,121,247,66,209,185,99,66,97,39,241,31,29,189,230,85,29,67,10,22,207,158,153,252,106,220,195,188,6,95,9,88,179,222,213,167,91,175,105,74,139,242,101,216,226,206,228,255,204,213,168,110,211,147,189,150,180,76,240,114,2,157,197,207,227,250,193,107,206,114,85,207,169,16,93,125,94,245,29,232,86,29,207,254,105,88,24,158,188,225,139,213,77,166,27,179,202,96,122,107,227,32,124,183,199,19,5,199,132,223,113,136,199,90,24,178,15,184,83,212,222,186,77,102,116,226,119,104,80,220,82,17,133,183,134,79,148,206,128,252,248,160,80,201,31,154,39,139,154,249,215,8,61,102,200,151,8,102,127,153,166,217,46,254,41,41,109,7,191,127,117,211,1,209,193,132,211,252,97,174,97,29,78,218,22,141,92,54,99,90,100,16,230,230,252,251,96,17,119,150,58,50,18,234,242,101,47,245,119,178,33,233,188,80,253,117,248,34,195,183,66,166,94,240,26,213,32,0,0,132,167,219,212,83,167,96,97,253,233,140,117,223,236,124,47,164,100,95,124,222,252,27,10,203,185,241,188,111,22,147,97,247,200,229,17,56,161,235,99,60,171,238,236,1,231,43,226,72,58,94,240,186,234,1,224,44,166,121,56,233,0,172,64,139,126,51,250,188,182,23,48,155,213,54,168,91,178,125,70,57,42,196,40,51,35,251,105,7,226,216,230,159,87,196,74,26,179,116,200,43,99,207,153,225,36,78,33,168,57,2,100,107,32,209,158,144,22,147,145,255,226,2,233,46,181,84,208,170,189,73,153,229,171,48,50,69,58,94,46,245,201,145,18,18,41,186,79,145,14,129,112,9,244,118,142,250,123,197,45,38,51,21,199,224,60,248,207,247,108,43,177,93,209,128,128,196,214,191,55,179,113,75,184,23,226,168,30,22,214,86,47,53,29,160,178,95,96,141,114,217,76,43,143,204,135,170,240,85,163,12,149,6,16,134,241,174,80,96,230,94,19,89,93,248,120,206,125,11,14,63,199,10,198,149,82,25,95,124,185,85,78,115,121,1,177,96,72,125,90,179,151,169,195,104,118,229,245,85,182,248,126,2,231,51,210,88,86,162,186,208,213,124,24,45,85,215,236,99,69,255,220,230,27,83,210,148,212,22,49,1,95,187,46,212,230,169,15,41,108,188,146,245,92,86,114,176,152,100,126,193,21,126,129,76,158,158,168],\"4\":[240,57,83,70,115,206,170,187,199,104,38,218,103,81,178,153,109,1,21,229,238,191,25,26,253,133,215,31,198,66,136,177,103,113,149,189,224,55,214,1,237,103,24,219,103,224,183,195,173,27,248,10,202,132,53,177,47,81,141,123,10,21,60,134,234,165,138,172,246,29,164,59,182,12,144,9,186,4,3,78,101,142,133,90,155,148,232,89,177,66,55,36,12,15,195,200,232,237,165,169,171,136,179,0,120,31,25,56,195,112,194,86,82,54,67,243,7,187,195,91,13,130,239,186,240,77,29,180,176,133,207,9,133,88,238,201,73,253,200,179,59,158,95,111,165,44,72,102,114,183,222,9,194,215,0,210,128,218,147,69,161,218,42,193,46,193,155,62,189,28,107,81,37,136,43,194,73,40,161,94,44,204,253,228,157,34,92,8,249,133,110,14,58,96,219,211,82,84,50,224,35,151,250,157,17,49,43,56,82,239,0,82,241,201,115,92,145,91,72,152,239,70,30,118,236,223,61,66,3,226,5,175,245,152,129,113,117,159,71,122,190,12,2,161,158,190,212,121,138,88,155,169,97,119,107,43,85,227,7,244,225,199,75,98,191,228,101,31,167,216,147,100,229,100,119,119,33,134,133,201,16,57,96,68,89,26,152,229,145,27,25,135,182,224,126,214,59,231,109,239,72,53,19,83,194,240,252,166,93,39,205,190,236,127,59,55,159,126,5,220,73,200,140,5,242,207,137,90,97,109,233,72,68,230,187,3,123,84,130,56,171,49,184,23,145,116,158,207,93,1,129,191,9,191,20,97,141,157,248,79,139,78,94,12,94,119,181,160,224,57,99,155,159,119,236,70,186,22,187,242,217,126,77,107,99,121,255,189,49,92,138,121,233,43,57,233,225,75,184,222,201,178,73,46,68,74,199,66,221,121,172,47,78,238,245,233,43,197,240,65,149,226,35,132,177,19,239,158,0,4,158,120,146,203,55,81,119,42,101,87,245,244,254,16,123,189,3,75,158,16,133,221,14,228,27,237,91,1,63,179,38,226,198,24,159,148,158,228,202,201,173,219,92,10,228,80,160,143,32,79,219,47,114,90,149,168,14,241,23,83,169,113,118,58,248,120,72,40,53,41,82,228,46,89,89,218,105,136,178,192,204,180,213,83,206,61,46,219,240,26,223,233,76,102,158,189,142,120,100,4,87,83,51,12,168,91,150,193,176,251,169,125,12,180,207,210,192,38,46,219,122,237,205,105,229,134,163,187,186,159,14,197,163,223,214,49,186,184,209,87,58,116,113,59,237,54,199,155,124,251,98,225,84,111,116,171,105,115,211,100,121,28,197,225,11,209,31,50,241,51,129,150,6,62,9,218,31,89,14,211,16,226,69,81,163,40,173,80,225,214,2,209,70,239,98,85,165,150,220,38,243,224,177,98,25,249,126,152,86,176,99,76,108,108,205,126,201,115,200,89,169,5,230,6,250,14,120,31,110,53,31,36,224,2,248,144,13,161,3,225,165,228,120,50,228,123,127,5,47,96,87,70,206,216,59,225,179,112,255,217,247,125,164,3,26,86,204,70,102,88,21,98,230,20,226,250,109,219,157,218,27,212,254,250,181,144,28,252,128,218,191,248,209,60,195,42,68,204,171,196,5,84,248,65,26,185,208,176,40,37,222,55,255,126,37,201,192,100,82,6,61,16,68,37,196,253,217,190,198,122,128,249,184,135,192,17,47,88,16,179,55,252,211,246,186,132,174,209,222,126,153,174,2,65,138,30,170,153,234,185,135,163,104,164,23,94,146,137,33,39,49,15,127,203,171,88,32,116,224,30,120,213,230,244,223,251,233,59,223,38,223,213,18,159,196,108,252,250,61,197,244,78,99,95,191,200,40,11,153,194,150,254,240,128,235,141,248,112,102,109,223,163,61,13,57,121,102,238,110,238,28,2,25,245,212,232,62,15,149,44,93,129,39,95,159,109,83,202,112,237,173,229,117,73,19,174,205,78,133,49,40,227,79,62,120,166,107,158,49,158,18,99,9,134,73,86,43,174,211,174,66,235,90,18,62,111,2,196,62,72,172,126,93,54,222,95,158,117,172,108,141,162,165,103,65,199,231,228,3,23,231,156,128,22,220,143,66,159,47,142,187,96,99,196,86,10,177,92,208,19,50,202,190,168,248,178,240,169,213,102,47,156,0,252,180,181,4,195,95,43,228,109,126,145,140,28,128,10,23,159,75,67,125,59,60,165],\"1\":[83,65,246,139,120,241,89,167,110,53,191,214,181,241,95,138,116,18,215,61,136,85,72,206,226,46,122,92,137,61,216,236,218,169,193,142,197,23,181,176,152,243,239,130,249,47,51,75,7,205,187,221,111,2,172,232,181,242,176,67,144,57,133,81,197,75,250,129,180,120,199,201,208,184,107,11,176,131,180,88,23,202,55,247,248,187,57,186,98,203,228,250,88,124,220,254,15,192,234,191,182,15,183,223,135,210,159,151,236,131,152,27,23,17,118,195,11,187,148,206,131,114,14,2,226,117,24,142,136,168,220,196,199,221,111,115,153,123,128,122,247,81,204,8,220,90,155,188,236,59,65,152,44,60,136,113,48,97,209,227,152,73,4,216,160,27,8,162,168,83,154,119,122,38,94,173,175,69,38,164,234,229,62,5,13,14,15,1,39,140,30,10,138,223,237,158,153,205,87,69,25,41,113,152,38,153,174,169,54,25,68,173,224,117,170,79,224,52,206,43,122,183,92,9,28,193,194,237,49,172,231,249,29,197,208,138,2,90,255,32,71,83,64,138,220,194,187,226,178,108,94,249,28,34,129,0,116,241,26,1,118,15,34,152,45,102,249,122,179,43,56,95,169,127,91,12,162,106,193,65,250,27,163,205,202,188,220,64,47,109,174,132,92,33,23,231,108,126,36,57,9,188,162,167,224,65,106,66,204,191,212,127,179,72,241,67,192,153,58,151,216,15,199,89,112,111,130,98,213,66,113,222,215,65,225,208,91,12,195,226,5,250,2,219,250,248,189,221,62,33,88,98,189,206,236,39,223,153,28,236,94,185,209,221,93,133,146,15,127,204,142,16,96,36,198,87,52,247,69,65,188,29,216,170,247,28,250,36,25,12,28,194,204,88,127,216,73,209,180,146,86,47,201,138,250,146,183,143,88,107,60,128,157,34,188,240,107,157,249,11,235,85,229,147,183,19,54,220,196,85,3,28,205,67,87,214,182,7,223,7,244,229,138,189,104,156,97,101,11,155,211,81,205,125,46,57,12,172,27,36,55,124,6,247,149,202,54,39,5,58,198,212,77,218,149,39,205,47,196,116,98,213,65,221,243,23,196,129,95,130,116,19,217,27,86,46,224,166,151,222,134,150,59,50,70,151,97,232,96,89,255,27,77,244,128,51,198,184,126,29,193,57,35,79,231,84,97,55,132,145,60,112,113,122,122,172,34,37,135,21,252,186,14,31,91,154,137,147,125,254,149,45,176,14,146,67,167,191,101,210,69,247,158,250,174,232,51,9,67,156,48,223,214,149,12,240,32,174,210,167,2,185,175,72,136,133,122,35,160,175,147,121,1,209,179,62,223,100,199,228,114,209,14,88,250,107,249,195,102,183,180,78,234,237,95,96,255,22,232,7,43,8,109,226,77,93,218,95,20,37,109,170,160,78,53,224,48,187,92,143,7,155,132,88,136,56,178,225,34,70,221,28,251,125,206,138,36,137,180,4,161,133,87,219,122,185,241,216,25,154,139,60,89,190,184,194,89,105,195,148,135,38,158,141,60,192,174,164,188,175,129,30,247,232,203,128,170,135,184,44,174,84,102,164,203,115,206,125,73,243,42,201,249,222,214,140,198,85,129,65,90,154,156,101,102,103,165,70,63,38,130,131,73,86,112,105,246,148,14,165,58,252,45,57,181,64,118,241,142,167,137,63,102,207,118,154,48,241,203,138,107,25,139,29,16,3,230,30,105,57,161,136,128,125,156,56,105,47,42,125,156,106,80,102,89,8,55,104,215,254,110,132,213,46,28,61,81,114,95,30,186,7,142,15,180,102,61,203,34,132,50,133,13,58,40,251,27,115,206,165,182,82,109,179,26,109,47,40,196,244,83,27,219,178,95,92,237,210,83,182,178,161,102,214,80,193,246,44,1,65,168,82,250,244,129,202,141,70,164,74,55,178,156,73,191,131,79,27,50,21,128,44,6,45,173,41,245,193,53,183,99,75,104,138,206,196,43,68,16,153,141,126,200,43,74,168,222,48,145,243,67,118,56,146,104,159,136,119,209,183,224,113,123,8,210,242,247,18,170,77,189,236,141,181,225,102,21,167,54,167,91,112,25,57,40,5,21,89,57,117,111,78,24,38,0,94,229,200,17,143,216,237,141,69,164,48,235,169,234,153,255,167,12,200,251,164,10,245,198,6,187,81,192,157,65,171,131,3,27,178,60,254,62,28,68,5,22,189,200,105,140,95,139,152,136,149],\"8\":[163,17,66,120,5,244,188,246,165,142,192,57,214,175,164,32,233,218,67,224,33,165,156,16,32,188,39,211,111,46,82,235,71,80,21,214,155,148,89,121,7,68,112,113,116,189,107,103,178,148,41,80,67,173,8,142,38,245,240,85,36,249,103,208,105,42,193,97,157,51,101,110,248,247,98,222,57,75,195,184,79,10,191,153,211,144,177,97,92,250,61,176,230,190,159,43,63,1,0,31,202,109,239,242,251,129,58,10,148,97,254,13,214,179,167,7,179,228,255,113,46,97,21,118,141,84,46,251,179,0,172,26,31,219,144,132,121,77,241,108,218,124,119,255,105,67,254,45,131,5,223,6,194,74,11,93,249,36,80,232,54,156,208,230,14,56,83,53,243,227,253,254,41,174,155,59,18,105,153,53,33,69,84,228,29,173,54,152,205,111,55,3,203,151,160,177,161,154,80,30,89,87,154,151,42,0,185,42,242,171,213,250,148,243,73,98,30,162,136,188,8,46,141,154,25,238,142,27,14,162,190,195,204,86,113,158,70,181,141,248,45,90,170,158,111,167,200,174,97,21,111,214,251,68,91,37,247,217,133,47,120,45,194,234,168,233,85,120,104,71,151,137,155,244,118,78,81,194,217,57,74,239,25,112,202,176,15,156,84,72,194,20,181,180,198,57,169,1,90,24,85,118,227,199,151,213,47,39,47,128,82,58,237,123,144,188,221,253,139,154,154,159,64,207,238,71,60,136,116,159,180,44,209,50,199,106,178,179,254,71,21,22,7,120,167,52,191,82,231,106,177,44,158,164,103,175,21,237,220,20,156,241,251,243,108,64,113,95,206,251,81,101,28,107,247,2,75,159,5,175,99,14,160,179,141,115,153,118,189,42,108,93,195,204,168,8,187,146,205,52,248,105,91,5,30,71,239,27,176,10,147,50,206,242,207,195,221,220,94,134,71,149,8,29,60,193,113,237,178,82,248,98,214,230,242,122,253,128,125,189,133,252,101,231,66,209,253,211,13,9,185,136,2,151,100,175,85,87,251,60,201,43,75,6,118,133,247,43,187,240,65,206,13,180,196,16,164,118,106,142,50,180,64,17,230,139,235,197,225,158,145,105,188,156,117,144,110,248,204,245,105,133,73,213,245,146,70,19,15,180,232,42,214,234,221,183,158,209,126,29,139,75,94,93,52,153,158,249,195,197,15,75,187,53,114,168,253,77,239,34,39,188,63,205,85,243,7,186,193,147,78,71,132,129,97,242,127,219,222,135,253,102,200,116,239,71,191,137,75,47,37,1,37,223,169,127,30,213,125,153,193,254,136,99,44,82,141,198,246,211,79,34,128,216,193,100,117,137,43,217,83,143,207,83,99,180,42,253,214,124,247,9,215,139,220,75,166,217,92,55,71,242,169,32,1,59,225,204,98,90,223,175,71,28,91,201,196,42,227,186,170,67,193,32,242,8,137,183,94,150,120,32,42,22,141,172,56,38,68,155,128,133,191,183,67,127,61,247,3,60,187,210,6,38,243,51,17,9,84,216,241,95,219,116,66,153,41,133,178,11,240,148,170,225,93,248,250,226,119,52,201,33,17,120,198,155,201,170,244,130,122,88,162,131,104,172,167,184,236,82,202,249,6,144,123,170,186,67,154,236,40,71,81,102,91,98,227,160,217,7,241,193,116,215,146,164,123,210,134,208,141,4,180,223,54,83,112,67,163,16,138,33,110,160,191,12,164,237,131,14,223,179,53,218,172,128,184,225,92,109,58,53,136,167,79,239,2,88,5,188,132,232,101,104,9,80,3,252,25,52,170,72,134,81,166,226,201,170,110,39,108,173,21,154,74,89,165,47,214,115,91,10,183,245,146,193,179,113,226,119,124,61,8,202,15,218,139,23,56,246,56,21,236,2,246,161,212,86,150,175,142,163,236,54,61,102,138,190,187,230,90,53,52,202,17,42,149,210,97,59,6,152,172,184,208,140,119,35,47,246,126,159,198,99,27,233,108,144,45,134,9,155,130,95,94,210,64,141,178,107,133,101,26,228,156,109,217,211,232,53,184,188,44,33,104,122,108,204,21,75,167,199,222,186,197,58,0,37,16,71,119,252,197,251,25,53,132,9,243,31,14,246,177,196,37,30,168,217,47,147,88,104,186,174,32,87,13,53,184,198,217,195,53,221,66,3,152,227,232,147,79,58,22,98,188,137,220,14,171,121,42,158,123,4,164,2,209,5,36,31,158,250,201,25,5,67],\"3\":[9,40,233,153,173,55,216,205,244,126,162,110,157,130,69,251,118,87,249,182,222,27,140,244,48,27,127,46,243,185,193,244,102,201,100,171,44,161,215,111,174,202,73,173,216,139,200,116,153,158,234,28,105,94,102,77,193,13,133,94,232,11,229,252,233,80,74,79,136,9,65,174,221,137,94,25,149,229,107,147,241,227,221,58,38,190,67,128,163,203,12,83,236,227,81,49,52,27,17,184,212,139,108,42,15,243,12,189,83,113,106,106,98,48,185,119,115,29,171,206,191,50,202,4,27,128,99,211,141,218,35,151,154,248,84,129,193,154,4,184,108,216,62,92,86,39,226,4,142,213,224,82,123,211,254,40,14,241,3,113,69,227,66,195,117,107,23,65,197,71,243,243,38,99,13,171,120,255,125,95,230,244,222,156,110,215,248,226,73,163,210,12,169,99,60,106,212,207,86,80,111,86,230,90,10,207,50,255,211,63,92,203,40,176,95,83,255,167,192,229,33,96,166,170,230,147,69,143,13,36,150,45,242,46,139,211,75,61,164,87,43,133,238,108,186,157,127,67,7,140,21,96,225,197,179,108,88,148,114,100,82,114,174,66,125,195,78,242,116,29,203,43,122,123,244,168,9,80,0,212,132,190,15,142,107,86,127,2,20,75,226,106,143,136,232,58,35,223,48,97,231,47,12,243,115,72,186,39,249,250,4,37,87,76,209,0,119,95,249,110,153,186,111,238,64,116,8,77,129,184,169,166,202,186,135,195,233,189,44,208,122,243,95,138,88,144,46,234,65,62,39,199,209,128,212,122,222,165,67,166,213,157,29,130,239,52,3,47,24,17,93,78,142,202,222,160,79,67,70,107,138,200,167,254,183,245,230,230,79,221,169,95,113,229,170,72,4,132,6,40,133,131,85,165,51,230,91,181,181,135,3,89,26,43,174,74,154,196,227,91,251,64,32,176,39,245,110,79,158,160,9,89,189,232,224,16,203,41,69,243,119,241,12,91,222,26,237,49,87,208,242,176,44,91,70,51,19,216,70,196,38,79,45,220,17,113,18,114,169,151,253,45,30,85,50,228,162,51,91,172,46,201,238,189,102,244,160,129,188,165,193,104,218,193,224,27,239,8,133,95,168,185,163,207,190,105,10,60,8,235,82,160,210,212,17,40,129,6,34,81,35,25,132,236,202,14,150,76,111,77,64,217,86,35,248,30,2,2,226,115,244,209,20,76,19,133,86,208,3,49,71,124,156,222,16,31,49,200,52,92,213,88,100,7,222,184,50,223,83,223,43,56,165,123,137,45,87,150,142,154,253,154,109,193,42,190,170,125,86,100,0,93,248,10,248,151,218,219,55,252,183,214,219,56,4,215,8,130,146,18,175,159,146,64,171,193,27,19,211,92,221,195,224,23,44,195,176,225,173,169,169,151,42,115,59,176,174,85,117,200,200,216,78,241,50,199,136,75,59,193,93,239,166,146,172,112,151,146,21,65,23,25,193,4,109,87,160,133,2,159,97,114,225,208,3,123,105,90,48,189,148,83,99,194,1,103,167,116,238,182,150,133,233,247,82,201,127,173,24,154,37,98,247,203,32,95,171,95,200,92,132,230,254,34,243,253,113,126,199,15,205,195,57,210,3,183,192,89,108,31,227,178,155,0,31,187,233,184,159,142,221,235,101,9,249,145,154,120,12,222,24,178,31,96,252,137,28,0,96,73,25,233,229,169,115,93,97,152,54,228,236,148,243,224,97,137,154,86,42,51,6,240,4,94,85,60,182,215,80,252,144,98,148,23,157,66,222,164,22,42,40,14,66,130,94,182,224,113,183,64,140,252,114,118,93,136,48,20,34,127,19,95,97,137,99,121,14,124,5,24,251,215,92,187,225,228,218,172,40,78,197,166,104,57,162,192,174,119,94,137,11,152,68,184,243,143,122,139,85,230,22,190,226,251,146,182,89,218,89,180,96,250,21,112,231,76,213,119,132,2,162,150,79,206,17,241,61,20,172,188,164,108,114,38,96,227,244,228,134,48,133,86,9,123,125,209,51,245,87,142,246,197,136,81,230,145,101,105,186,154,189,82,244,177,228,63,88,78,139,109,157,71,248,180,7,211,192,242,99,118,176,103,189,225,31,167,96,231,120,243,162,161,6,175,155,59,31,148,86,137,18,97,213,170,80,254,67,68,228,47,59,195,99,46,86,227,99,133,42,52,6,80,48,192,135,118,83,200,127,99,101,165,97,153,27,188],\"6\":[90,149,140,53,121,148,108,180,117,11,173,24,107,54,142,171,218,60,71,105,217,215,43,46,173,123,191,107,122,83,66,74,253,162,177,49,80,137,8,35,213,158,73,223,188,123,59,201,123,174,197,146,239,180,13,217,165,229,173,243,106,162,100,177,1,14,59,7,47,76,29,116,141,60,194,240,73,224,215,233,186,21,191,4,201,128,204,182,157,240,55,244,80,184,209,247,147,96,177,231,215,11,75,148,176,242,112,67,169,38,94,30,133,37,184,98,184,58,123,227,189,112,98,198,128,212,41,23,68,243,236,117,174,69,20,97,48,168,128,254,239,203,132,172,190,92,230,154,40,213,88,131,131,56,159,119,154,125,114,228,200,121,81,155,195,164,47,236,1,67,34,253,191,80,179,143,19,91,31,1,38,104,233,193,205,135,2,243,31,132,228,206,40,224,178,228,206,115,84,87,172,81,10,169,230,171,58,128,15,98,150,231,84,233,198,41,21,23,68,17,242,222,46,142,246,177,46,250,72,108,21,55,129,136,105,168,199,12,19,44,49,217,222,203,89,19,98,191,229,48,145,132,109,44,96,23,223,11,146,16,43,30,55,118,24,222,52,255,223,158,113,105,0,111,80,140,236,34,90,102,177,175,161,154,218,185,139,253,227,105,53,178,242,136,30,199,77,103,100,187,68,187,112,221,113,255,15,205,83,237,111,56,5,121,65,205,100,102,172,63,228,29,201,41,64,99,226,235,69,42,186,161,131,186,231,39,125,244,245,74,168,16,218,179,247,182,26,156,49,191,85,196,226,16,8,104,82,19,64,52,174,200,62,56,76,190,94,182,32,61,168,223,131,147,146,227,213,205,31,16,169,70,7,68,175,4,172,45,129,11,83,47,228,62,97,144,197,64,97,7,254,152,244,33,102,73,137,50,75,98,203,250,228,208,20,205,207,52,69,45,223,58,71,128,55,116,132,58,97,201,235,116,198,25,157,226,89,188,52,28,167,232,217,219,157,150,132,175,136,37,28,104,119,42,170,138,68,155,216,236,119,200,205,16,84,253,178,3,178,25,239,250,136,214,109,237,185,216,245,166,209,14,161,220,233,238,95,38,144,29,42,72,245,85,204,230,186,102,149,41,208,35,207,71,163,130,18,11,65,15,36,253,44,126,247,209,2,39,46,142,45,215,131,226,58,104,236,144,107,58,252,224,105,125,251,191,104,13,200,240,157,154,58,207,73,219,235,131,18,202,186,66,201,190,3,0,45,82,232,23,243,7,154,243,128,175,247,240,172,131,114,213,57,70,254,3,44,226,19,174,51,81,69,58,194,170,244,161,187,239,238,37,29,175,237,196,234,252,179,107,233,237,39,93,68,155,75,234,45,187,209,165,153,151,35,11,247,98,94,16,86,117,4,90,109,27,198,128,2,44,253,191,107,76,167,84,137,64,197,122,168,18,20,97,28,128,241,41,6,94,134,25,66,96,215,245,134,203,174,215,32,80,62,22,116,226,145,2,244,165,171,103,12,71,226,213,59,2,246,143,196,252,124,70,104,136,227,185,244,39,126,7,37,201,117,106,107,90,157,195,134,140,220,31,240,42,114,123,207,123,204,69,226,13,93,31,250,142,108,63,92,132,136,16,70,241,34,16,24,206,139,218,133,140,1,200,193,150,88,3,153,109,196,50,3,233,230,228,116,221,133,195,104,67,95,27,2,141,19,35,28,205,176,163,6,119,208,11,104,57,233,41,50,3,93,135,150,147,174,130,131,34,14,167,175,155,244,184,71,96,36,63,148,39,45,119,20,182,97,143,44,237,184,53,200,7,225,217,233,23,94,129,225,206,36,229,85,130,139,57,212,130,101,182,78,205,201,43,29,23,18,183,176,108,134,52,180,219,67,166,10,85,36,123,9,13,89,237,65,6,58,51,117,133,123,4,1,190,159,197,9,140,217,204,239,2,43,231,70,196,114,153,231,17,254,19,80,75,61,218,142,66,157,71,172,94,8,32,7,133,222,166,32,34,126,225,230,163,128,124,199,118,149,212,109,117,39,255,133,163,171,33,175,194,233,143,224,114,237,34,33,151,250,210,213,118,135,57,4,177,51,204,39,81,248,239,253,102,249,199,110,102,241,170,30,29,93,163,77,192,27,71,132,225,108,66,5,64,20,148,63,66,121,126,136,126,44,82,188,221,155,250,151,101,32,97,218,213,126,152,139,88,3,53,177,220,10,91,205,108,54,208,79,252,23,23],\"2\":[35,58,103,185,133,5,217,87,173,98,13,139,28,66,70,181,216,126,92,223,91,190,162,214,239,23,164,194,176,197,187,46,223,24,241,122,159,20,161,75,168,116,153,227,60,9,164,69,215,225,21,15,121,50,18,42,105,0,140,158,59,18,48,161,182,220,105,162,6,60,90,88,80,17,172,255,170,59,81,90,221,141,228,102,77,42,120,203,127,233,67,19,71,188,64,46,225,20,69,228,0,206,104,12,140,232,183,201,52,236,145,173,45,46,29,211,183,16,91,153,47,48,184,93,12,232,33,231,250,248,204,69,3,211,4,100,171,68,207,74,121,146,190,83,251,66,148,116,160,35,114,124,68,225,167,23,132,159,253,71,68,235,198,117,96,69,96,51,85,81,209,96,152,94,102,11,195,75,105,16,134,238,218,103,242,82,47,23,9,172,51,53,2,193,16,169,105,211,112,33,198,167,42,68,79,14,1,167,88,172,194,52,89,244,140,160,47,182,212,237,165,57,209,233,52,208,154,231,86,116,48,30,76,19,7,20,124,94,202,187,253,133,167,92,81,29,142,141,120,44,126,39,6,24,54,180,248,189,110,200,175,178,254,135,137,167,84,10,206,109,66,20,230,186,46,190,91,170,167,216,187,129,217,34,184,224,225,216,102,144,105,137,16,244,188,254,130,53,108,176,126,166,2,207,182,252,60,15,97,8,69,248,58,92,184,25,19,186,18,77,227,140,17,248,169,101,134,109,1,102,238,197,118,232,234,12,70,212,99,83,223,126,226,55,8,18,228,67,65,121,44,90,48,97,195,244,109,0,55,170,19,192,242,180,148,84,200,101,86,9,237,188,12,171,166,184,27,202,144,53,44,126,129,126,198,143,37,107,51,202,25,20,164,49,36,126,68,38,32,12,224,203,203,67,196,43,139,98,246,226,226,114,28,53,16,105,79,163,40,212,21,228,3,20,20,218,29,98,57,64,117,127,99,201,143,101,78,19,7,4,252,131,109,27,189,204,27,10,96,174,53,126,251,10,164,19,21,81,182,40,123,53,60,227,133,120,108,26,149,25,77,67,125,118,83,18,164,118,94,248,206,224,125,103,188,184,51,109,80,118,99,86,194,78,163,32,45,64,106,174,58,45,0,75,54,67,118,236,171,113,12,26,148,77,57,146,66,12,245,53,172,25,214,205,252,112,230,100,127,32,217,104,124,183,88,237,175,175,88,126,87,194,208,190,89,219,67,206,151,88,217,112,56,119,57,167,1,250,239,190,196,136,21,225,244,236,119,28,253,204,243,241,56,99,163,199,186,254,41,119,103,102,241,208,152,237,210,161,32,93,35,238,213,255,208,33,201,172,74,73,224,172,133,216,145,30,208,186,222,81,192,151,183,154,102,233,30,3,95,196,41,89,248,132,255,101,5,143,151,14,122,189,219,115,234,104,191,163,4,190,129,129,200,90,155,111,113,213,125,170,12,138,185,23,110,149,41,135,209,122,136,87,239,115,204,198,170,246,79,218,227,144,148,23,199,235,214,1,58,44,105,49,151,140,80,37,249,124,113,59,204,203,134,34,179,94,231,149,142,126,12,232,168,214,114,11,62,39,59,96,51,190,5,128,184,92,163,177,99,130,6,89,114,221,40,175,174,5,177,112,160,28,196,20,3,32,68,221,181,104,92,250,191,89,97,29,241,225,107,144,235,212,173,120,71,21,213,185,12,231,154,58,20,151,205,122,175,220,198,131,18,68,91,207,83,248,148,165,28,135,122,205,213,202,227,226,121,121,182,144,191,146,97,158,128,11,119,233,128,106,195,128,105,42,171,55,179,9,5,210,118,50,225,38,171,197,14,97,11,131,27,250,67,107,95,147,53,1,93,62,153,199,149,95,22,200,222,143,80,70,253,205,111,5,245,59,105,254,206,251,238,182,120,56,94,160,71,120,57,28,51,199,215,93,10,220,118,87,38,148,10,244,254,37,114,130,236,122,86,88,62,194,26,101,31,194,226,46,26,90,75,167,185,221,146,117,47,152,133,140,178,17,52,253,142,174,168,223,87,111,132,225,105,192,181,169,7,186,14,19,236,152,11,160,143,4,129,181,115,88,61,156,3,194,84,7,222,137,196,192,168,117,122,160,132,131,103,113,59,129,112,35,8,229,111,44,66,176,252,64,34,243,76,30,146,116,232,47,166,100,246,215,36,12,145,229,129,253,29,6,129,151,28,172,205,109,128,77,115,9,180,145,226,7,211,100]},\"reachable\":{\"{\\\"addr\\\":[49,57,49,46,49,56,48,46,49,48,51,46,50,51,54,0],\\\"port\\\":50309,\\\"uid_fingerprint\\\":17576062521251310994,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[6,0],[7,0]],\"{\\\"addr\\\":[49,51,50,46,49,56,49,46,49,53,53,46,55,54,0,0],\\\"port\\\":55241,\\\"uid_fingerprint\\\":4215620865531854407,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,67,66,70,56,51,51,51,49,54,53,70,49,67,51,53,65,68,70,48,54,66,70,52,50,53,49,48,52,69,56,51,51,65,68,68,65,49,54,53,66,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[1,1]],\"{\\\"addr\\\":[49,53,54,46,55,50,46,50,49,57,46,52,57,0,0,0],\\\"port\\\":56030,\\\"uid_fingerprint\\\":8564141605929456761,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,68,56,66,67,67,48,69,67,70,69,50,67,57,54,53,65,66,67,50,57,67,54,50,55,70,68,67,69,67,48,66,48,54,67,53,65,68,66,49,54,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[1,2]],\"{\\\"addr\\\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\\\"port\\\":63174,\\\"uid_fingerprint\\\":922068108643293997,\\\"info\\\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,70,68,56,68,67,55,69,70,57,50,70,49,70,49,52,68,48,48,67,70,57,68,54,70,54,50,57,55,65,51,52,54,56,66,53,57,69,55,48,55,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]}\":[[6,1],[8,0]],\"{\\\"addr\\\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\\\"port\\\":63174,\\\"uid_fingerprint\\\":9189545078772697644,\\\"info\\\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,69,68,67,51,65,65,48,51,70,69,54,69,53,50,66,69,55,55,56,68,68,52,54,70,50,55,68,51,53,56,57,57,51,66,55,65,49,48,65,52,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]}\":[[6,2],[9,0]],\"{\\\"addr\\\":[50,49,54,46,49,49,55,46,51,46,54,50,0,0,0,0],\\\"port\\\":63174,\\\"uid_fingerprint\\\":14821306939685929169,\\\"info\\\":[116,121,112,101,61,115,99,114,97,109,98,108,101,115,117,105,116,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,66,48,56,49,49,56,65,56,48,68,55,54,54,66,65,57,65,56,70,57,65,55,50,50,67,50,54,54,49,52,69,48,54,66,56,54,69,52,67,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,83,111,109,101,40,123,34,112,97,115,115,119,111,114,100,34,58,32,34,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,50,51,52,53,54,55,34,125,41,0]}\":[[10,0]],\"{\\\"addr\\\":[49,51,52,46,50,48,55,46,50,51,53,46,49,48,53,0],\\\"port\\\":9042,\\\"uid_fingerprint\\\":5838789323103363859,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,56,48,55,56,70,57,55,66,50,50,49,53,55,69,66,55,49,54,65,69,66,68,53,54,65,57,52,55,54,67,51,55,56,57,57,65,68,70,54,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[2,2],[5,0]],\"{\\\"addr\\\":[50,50,46,50,52,49,46,56,49,46,54,50,0,0,0,0],\\\"port\\\":36457,\\\"uid_fingerprint\\\":14409951350199027285,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,68,68,56,56,57,53,48,56,48,48,50,70,57,50,53,67,70,55,52,69,49,51,48,68,49,48,49,67,68,70,49,67,49,51,51,65,65,65,68,49,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[10,2]],\"{\\\"addr\\\":[49,57,53,46,55,50,46,55,52,46,49,49,55,0,0,0],\\\"port\\\":28746,\\\"uid_fingerprint\\\":8013544061081454931,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,50,57,49,49,48,66,55,68,57,69,57,54,48,65,53,67,52,50,56,52,56,55,55,50,65,70,54,50,52,65,51,66,50,55,65,50,68,69,52,69,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[2,0],[3,0]],\"{\\\"addr\\\":[49,50,56,46,49,51,57,46,52,52,46,49,53,55,0,0],\\\"port\\\":10839,\\\"uid_fingerprint\\\":16072135124933872676,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,55,67,53,65,51,68,67,53,51,49,48,55,53,53,48,54,65,50,52,52,65,51,67,70,49,57,66,53,55,70,52,68,69,52,55,55,67,65,54,51,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[2,1],[4,0]],\"{\\\"addr\\\":[50,49,50,46,49,54,54,46,49,54,54,46,54,55,0,0],\\\"port\\\":32414,\\\"uid_fingerprint\\\":3643046579853129326,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,66,48,56,49,49,56,65,56,48,68,55,54,54,66,65,57,65,56,70,57,65,55,50,50,67,50,54,54,49,52,69,48,54,66,56,54,69,52,67,65,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,102,97,115,116,34,58,32,116,114,117,101,44,32,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[1,0]],\"{\\\"addr\\\":[57,55,46,49,55,53,46,57,49,46,55,0,0,0,0,0],\\\"port\\\":27719,\\\"uid_fingerprint\\\":9270304952963857555,\\\"info\\\":[116,121,112,101,61,111,98,102,115,50,32,98,108,111,99,107,101,100,95,105,110,61,123,125,32,112,114,111,116,111,99,111,108,61,116,99,112,32,102,105,110,103,101,114,112,114,105,110,116,61,34,69,68,67,51,65,65,48,51,70,69,54,69,53,50,66,69,55,55,56,68,68,52,54,70,50,55,68,51,53,56,57,57,51,66,55,65,49,48,65,52,34,32,111,114,95,97,100,100,114,101,115,115,101,115,61,78,111,110,101,32,100,105,115,116,114,105,98,117,116,105,111,110,61,104,116,116,112,115,32,102,108,97,103,115,61,83,111,109,101,40,123,34,115,116,97,98,108,101,34,58,32,116,114,117,101,44,32,34,114,117,110,110,105,110,103,34,58,32,116,114,117,101,44,32,34,118,97,108,105,100,34,58,32,116,114,117,101,44,32,34,102,97,115,116,34,58,32,116,114,117,101,125,41,32,112,97,114,97,109,115,61,78,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}\":[[10,1]]},\"spares\":[10,1],\"unallocated_bridges\":[],\"recycleable_keys\":[],\"blocked_keys\":[],\"open_inv_keys\":[[3,2460237],[4,2460237],[5,2460237],[7,2460237],[8,2460237],[9,2460237]],\"date_last_enc\":2460237},\"trustup_migration_table\":{\"table\":{\"5\":2,\"4\":2,\"9\":6,\"7\":6,\"3\":2,\"8\":6},\"migration_type\":[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},\"blockage_migration_table\":{\"table\":{},\"migration_type\":[1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]},\"openinv_filter\":{\"seen_table\":[[167,121,14,179,249,174,210,194,223,78,92,229,191,6,227,23,111,17,206,87,132,3,0,86,116,37,13,177,246,57,9,0],[20,146,141,79,128,206,56,95,212,130,37,81,177,11,14,184,9,71,14,116,101,107,243,17,237,199,192,127,212,63,30,7]]},\"id_filter\":{\"seen_table\":[]},\"inv_id_filter\":{\"seen_table\":[]},\"trust_promotion_filter\":{\"seen_table\":[]}},\"extra_bridges\":[],\"to_be_replaced_bridges\":[]}";
        let test_context: LoxServerContext = serde_json::from_str(&test_string).unwrap();
        lox_db.write_context(test_context);
        assert!(
            lox_db.db.len() == 1,
            "db should have written only one context"
        );
    }
}
